@@ -1,6 +1,7 @@
 # -*- coding: utf-8 -*-
 import datetime
 import numbers
+from typing import Any, Optional
 
 # from builtins import str
 
@@ -15,15 +16,16 @@ FM_TIMESTAMP = 'timestamp'
 
 class TypeCast:
     """Type caster, get's initiated with the corresponding FmFieldData"""
+
     def __init__(self, fm_field, fm_meta):
         pass
 
-    def __call__(self, value):
+    def __call__(self, value: str) -> Any:
         return value
 
 
 class NumberCast(TypeCast):
-    def __call__(self, value):
+    def __call__(self, value: str) -> float:
         try:
             return float(value)
         except Exception:
@@ -31,8 +33,17 @@ class NumberCast(TypeCast):
             return float('nan')
 
 
+class CommaDecimalNumberCast(TypeCast):
+    def __call__(self, value: str) -> float:
+        try:
+            return float(value.replace(',', '.'))
+        except Exception:
+            # return NaN
+            return float('nan')
+
+
 class TextCast(TypeCast):
-    def __call__(self, value):
+    def __call__(self, value: str) -> str:
         if value:
             return value
         return ''
@@ -45,7 +56,7 @@ class DateCast(TypeCast):
     def __init__(self, fm_field, fm_meta):
         self.pat = fm_meta.date_pattern
 
-    def __call__(self, value):
+    def __call__(self, value: str) -> Optional[datetime.date]:
         try:
             d = datetime.datetime.strptime(
                 value,
@@ -60,7 +71,7 @@ class TimeCast(TypeCast):
     def __init__(self, fm_field, fm_meta):
         self.pat = fm_meta.time_pattern
 
-    def __call__(self, value):
+    def __call__(self, value: str) -> Optional[datetime.time]:
         try:
             return datetime.datetime.strptime(
                 value,
@@ -75,15 +86,14 @@ class TimestampCast(TypeCast):
         self.pat = fm_meta.timestamp_pattern
         self.tz = fm_meta.server_timezone
 
-    def __call__(self, value):
+    def __call__(self, value: str) -> Optional[datetime.datetime]:
         try:
             d = datetime.datetime.strptime(
                 value,
                 self.pat,
             )
             if self.tz:
-                d = self.tz.localize(d)
-                d = self.tz.normalize(d)
+                d = d.replace(tzinfo=self.tz)
             return d
         except (ValueError, TypeError):
             return None
@@ -96,36 +106,40 @@ class BackCast:
     FM_DEFAULT_TIMESTAMP = "%m/%d/%Y %H:%M:%S"
 
     def __init__(self, fm_server=None):
-        """The :fm_server: object is passed at the initialisation of this class.
-        It can be used to cast some field in a different way"""
+        """
+        The :fm_server: object is passed at the initialisation of this class.
+        It can be used to cast some field in a different way
+        """
         if fm_server:
             self.tz = fm_server.options['server_timezone']
 
-    def __call__( self, field, value ):
-        if isinstance( value, datetime.datetime ):
+    def __call__(self, field, value):
+        if isinstance(value, datetime.datetime):
             # if server timezone is set and the datetime is aware:
-            if self.tz and value.tzinfo is not None and value.tzinfo.utcoffset(value) is not None:
+            if (
+                self.tz and
+                value.tzinfo is not None and
+                value.tzinfo.utcoffset(value) is not None
+            ):
                 if self.tz != value.tzinfo:
                     value = value.astimezone(self.tz)
-            return value.strftime( self.__class__.FM_DEFAULT_TIMESTAMP )
+            return value.strftime(self.__class__.FM_DEFAULT_TIMESTAMP)
 
-        elif isinstance( value, datetime.date ):
-            return value.strftime( self.__class__.FM_DEFAULT_DATE )
+        elif isinstance(value, datetime.date):
+            return value.strftime(self.__class__.FM_DEFAULT_DATE)
 
-        elif isinstance( value, datetime.time ):
-            return value.strftime( self.__class__.FM_DEFAULT_TIME )
+        elif isinstance(value, datetime.time):
+            return value.strftime(self.__class__.FM_DEFAULT_TIME)
 
-        elif isinstance( value, bytes ):
+        elif isinstance(value, bytes):
             return value.decode('utf8')
 
-        elif isinstance( value, numbers.Number ):
+        elif isinstance(value, numbers.Number):
             return value
 
         return str(value)
 
 
-
-
 default_cast_map = {
     FM_NUMBER: NumberCast,
     FM_TEXT: TextCast,
