@@ -9,7 +9,14 @@ from pyfilemaker2.caster import BackCast
 class TestBackCast(unittest.TestCase):
     def test_backcast_timezone(self):
         tz = pytz.timezone('Europe/Zurich')
-        d = datetime.datetime(year=2018, month=9, day=1, hour=11, minute=54, second=7)
+        d = datetime.datetime(
+            year=2018,
+            month=9,
+            day=1,
+            hour=11,
+            minute=54,
+            second=7
+        )
         d = tz.normalize(tz.localize(d))
 
         tz2 = pytz.UTC
@@ -20,7 +27,14 @@ class TestBackCast(unittest.TestCase):
 
     def test_backcast_timezone2(self):
         tz2 = pytz.UTC
-        d = datetime.datetime(year=2018, month=9, day=1, hour=11, minute=54, second=7)
+        d = datetime.datetime(
+            year=2018,
+            month=9,
+            day=1,
+            hour=11,
+            minute=54,
+            second=7
+        )
         d = tz2.normalize(tz2.localize(d))
 
         tz = pytz.timezone('Europe/Zurich')
@@ -29,7 +43,14 @@ class TestBackCast(unittest.TestCase):
         self.assertEqual(bc(field=None, value=d), '09/01/2018 13:54:07')
 
     def test_backcast_timezone3(self):
-        d = datetime.datetime(year=2018, month=9, day=1, hour=11, minute=54, second=7)
+        d = datetime.datetime(
+            year=2018,
+            month=9,
+            day=1,
+            hour=11,
+            minute=54,
+            second=7
+        )
 
         tz = pytz.timezone('Europe/Zurich')
         bc = BackCast()
